#include <stdio.h>
#include <string.h>

#define HELP1 "/?"
#define HELP2 "-?"

#define STRINF "<"
#define INF "<="
#define EGAL1 "="
#define EGAL2 "=="
#define SUP ">="
#define STRSUP ">"
#define DIFF1 "<>"
#define DIFF2 "!="

#define NOERROR 0
#define ARGERROR 1
#define IOERROR 2
#define INTERROR 4

#define elsif else if

#define print_openerror(F)\
fprintf(stderr,"Impossible d'ouvrir le fichier %s\n",F)
#define print_rderror(F) fprintf(stderr,"Impossible de lire le fichier %s\n",F)
#define print_wrerror(F)\
fprintf(stderr,"Impossible d'ecrire dans le fichier %s\n",F)
#define print_interror() fprintf(stderr,"Erreur interne critique !\n")

/* ------------------------- Comparaisons possibles ------------------------- */

enum comp {
  no_comp = 0,
  str_inf = 1,
  inf = 2,
  egal = 3,
  sup = 4,
  str_sup = 5,
  diff = 6
};

typedef enum comp comp;

/* ---------------------------- Actions possibles --------------------------- */

enum action {
  no_action = 0,
  aide = 1,
  egal_val = 2,
  comp_val = 3,
  egal_fich = 4,
  diff_fich = 5,
  comp_fich = 6
};

typedef enum action action;

/* ----------- Tester si une chaine est un nombre et la convertir ----------- */

int str_int(const char *str, int min, int max, int *val)
{
  char u;

  return(sscanf(str,"%d%c",val,&u) == 1 && *val <= max && *val >= min);
}

/* ---------- Tester si une chaine est un operateur de comparaison ---------- */

comp str_comp(const char *str)
{
  if (!strcmp(str,STRINF))
    return(str_inf);
  elsif (!strcmp(str,INF))
    return(inf);
  elsif (!strcmp(str,EGAL1) || !strcmp(str,EGAL2))
    return(egal);
  elsif (!strcmp(str,SUP))
    return(sup);
  elsif (!strcmp(str,STRSUP))
    return(str_sup);
  elsif (!strcmp(str,DIFF1) || !strcmp(str,DIFF2))
    return(diff);
  else
    return(no_comp);
}

/* ------------------ Appliquer un operateur de comparaison ----------------- */

int comparer(int v1, comp op, int v2)
{
  switch (op)
  {
    case str_inf :
      return(v1 < v2);
    break;
    case inf :
      return(v1 <= v2);
    break;
    case egal :
      return(v1 == v2);
    break;
    case sup :
      return(v1 >= v2);
    break;
    case str_sup :
      return(v1 > v2);
    break;
    case diff :
      return(v1 != v2);
    break;
    default :
      print_interror();
      return(0);
  }
}

/* ------------------------- Parametres du programme ------------------------ */

struct st_data {
  action a;                              /* action demandee par l'utilisateur */
  comp c;                            /* comparaison a effectuer s'il y a lieu */
  int v;                                    /* valeur/difference a rechercher */
  char *f1, *f2;                                   /* noms des fichier a lire */
  char *sc;                                        /* nom du fichier resultat */
};

typedef struct st_data *pdata;

/* ------------------ Lire et analyser la ligne de commande ----------------- */

void param(int argc, char *argv[], pdata data)
{
  comp op;
  int val;

  if (argc == 2 && (!strcmp(argv[1],HELP1) || !strcmp(argv[1],HELP2)))
    data->a = aide;
  elsif (argc == 4 && str_int(argv[2],0,255,&val))
  {
    data->a = egal_val;
    data->c = egal;
    data->v = val;
    data->f1 = argv[1];
    data->f2 = NULL;
    data->sc = argv[3];
  }
  elsif (argc == 5 && (op = str_comp(argv[2])) && str_int(argv[3],0,255,&val))
  {
    data->a = comp_val;
    data->c = op;
    data->v = val;
    data->f1 = argv[1];
    data->f2 = NULL;
    data->sc = argv[4];
  }
  elsif (argc == 4)
  {
    data->a = egal_fich;
    data->c = egal;
    data->v = 0;
    data->f1 = argv[1];
    data->f2 = argv[2];
    data->sc = argv[3];
  }
  elsif (argc == 5 && str_int(argv[2],-255,255,&val))
  {
    data->a = diff_fich;
    data->c = no_comp;
    data->v = (unsigned char) val;
    data->f1 = argv[1];
    data->f2 = argv[3];
    data->sc = argv[4];
  }
  elsif (argc == 5 && (op = str_comp(argv[2])))
  {
    data->a = comp_fich;
    data->c = op;
    data->v = 0;
    data->f1 = argv[1];
    data->f2 = argv[3];
    data->sc = argv[4];
  }
  else
    data->a = no_action;
}

/* ----------------------------- Message d'aide ----------------------------- */

void print_aide(void)
{
  printf("Pour rechercher un octet donnee dans un fichier :\n");
  printf("Scan file v result\n");
  printf("\n");
  printf("Pour rechercher les octets verifiant un test dans un fichier :\n");
  printf("Scan file op v result\n");
  printf("\n");
  printf("Pour rechercher les octets identiques dans deux fichiers :\n");
  printf("Scan file1 file2 result\n");
  printf("\n");
  printf(
 "Pour rechercher les octets ayant une difference donnee dans deux fichiers\n");
  printf("Scan file1 diff file2 result\n");
  printf("\n");
  printf("Pour rechercher les octets verifiant un test entre deux fichiers\n");
  printf("Scan file1 op file2 result\n");
  printf("\n");
  printf("  file         nom du fichier a scanner.\n");
  printf(
 "  file1 file2  noms des fichiers a comparer. Doivent avoir meme taille.\n");
  printf(
 "  v            octet a rechercher/tester. Doit valoir entre 0 et 255\n");
  printf(
 "  op           operateur de comparaison : <, >, <=, >=, =, ==, <> ou !=\n");
  printf(
 "  diff         difference entre octets a rechercher (entre -255 et 255).\n");
  printf("  result       fichier resultat. S'il n'exite pas il sera cree et\n");
  printf("               contiendra des 0 (faux) et des 1 (vrai).\n");
  printf(
 "               S'il existe il devra avoir meme taille que les autres \n");
  printf("               fichiers, et un ET logique sera applique.\n");

}

/* ------------------------ Comparaison des fichiers ------------------------ */

int comp_fichiers(pdata data, FILE *f1, FILE *f2, FILE *fsc, int new)
{
  int rd1, rd2, wr;
  long l = 0;

  do
  {
    if (new)
      wr = 1;
    else
      wr = fgetc(fsc);

    if (wr > 0 && wr != EOF)
    {
      if (l > 0)
      {
        if (fseek(f1,l,SEEK_CUR))
        {
          print_rderror(data->f1);
          return(IOERROR);
        }

        if (f2 != NULL && fseek(f2,l,SEEK_CUR))
        {
          print_rderror(data->f2);
          return(IOERROR);
        }

        l = 0;
      }

      rd1 = fgetc(f1);

      if (rd1 != EOF)
      {
        if (f2 == NULL)
          rd2 = 0;
        elsif ((rd2 = fgetc(f2)) == EOF)
        {
          print_rderror(data->f2);
          return(IOERROR);
        }

        switch (data->a)
        {
          case egal_val :
            wr = rd1 == data->v;
          break;
          case comp_val :
            wr = comparer(rd1,data->c,data->v);
          break;
          case egal_fich :
            wr = rd1 == rd2;
          break;
          case diff_fich :
            wr = data->v == (unsigned char)(rd1 - rd2);
          break;
          case comp_fich :
            wr = comparer(rd1,data->c,rd2);
          break;
          default :
            print_interror();
            return(INTERROR);
        }

        if (wr == 0 || new)
        {
          if ((!new && fseek(fsc,-1,SEEK_CUR))
           || fputc(wr,fsc) != wr
           || (!new && fseek(fsc,0,SEEK_CUR)))
          {
            print_wrerror(data->sc);
            return(IOERROR);
          }
        }
      }
      else /* rd1 == EOF */
      {
        if (!new)
        {
          print_rderror(data->f1);
          return(IOERROR);
        }
      }
    }
    else /* wr == 0 */
      l++;
  } while (wr != EOF && rd1 != EOF);

  return(NOERROR);
}

/* ------------------------- Ouverture des fichiers ------------------------- */

int run(pdata data)
{
  int new, res;
  FILE *f1, *f2, *fsc;

  if ((f1 = fopen(data->f1,"rb")) == NULL)
  {
    print_openerror(data->f1);
    return(IOERROR);
  }

  if (data->f2 == NULL)
    f2 = NULL;
  else
  {
    if ((f2 = fopen(data->f2,"rb")) == NULL)
    {
      print_openerror(data->f2);
      fclose(f1);
      return(IOERROR);
    }
  }

  if ((fsc = fopen(data->sc,"rb+")) != NULL)
  {
    new = 0;
    fseek(fsc,0,SEEK_SET);
  }
  else
  {
    new = 1;
    if ((fsc = fopen(data->sc,"wb")) == NULL)
    {
      fprintf(stderr,"Impossible de creer le fichier %s\n",data->sc);
      fclose(f1);
      if (f2 != NULL)
        fclose(f2);
      return(IOERROR);
    }
  }

  res = comp_fichiers(data,f1,f2,fsc,new);

  fclose(f1);
  if (f2 != NULL)
    fclose(f2);
  fclose(fsc);

  return(res);
}

/* --------------------------- Programme principal -------------------------- */

int main(int argc, char *argv[])
{
  int res;
  struct st_data data;

  param(argc,argv,&data);

  if (!data.a)
  {
    fprintf(stderr,"Arguments incorrects.\n");
    fprintf(stderr,"Pour l'aide : Scan /?\n");
    return(ARGERROR);
  }

  switch (data.a)
  {
    case aide :
      print_aide();
      res = NOERROR;
    break;
    case egal_val :
    case comp_val :
    case egal_fich :
    case diff_fich :
    case comp_fich :
      res = run(&data);
    break;
    default :
      print_interror();
      res = INTERROR;
    break;
  }

  return(res);
}
